// From http://www.redblobgames.com/grids/hexagons/
// Copyright 2015 Red Blob Games <redblobgames@gmail.com>
// License: Apache v2.0 <http://www.apache.org/licenses/LICENSE-2.0.html>
//
// I parse Haxe source into an abstract syntax tree, which I then
// convert into Python, C++, etc. Since I'm using macros, what I'm
// actually doing is at compile time converting a Haxe abstract syntax
// tree containing expressions such as A+B into another Haxe abstract
// syntax tree for OperatorPlus(Name("A"), Name("B")) that generates
// *my* abstract syntax tree at run time. Then at run time I convert
// that syntax tree into language-specific output strings.
 
import haxe.macro.Expr;

// Types
enum TExpr {
    TVoid;
    TInt;
    TFloat;
    TString;
    THex;
    TFractionalHex;
    TOffsetCoord;
    TDoubledCoord;
    TPoint;
    TLayout;
    TOrientation;
    TArray(t:TExpr);
}

// Expressions and statements
enum AExpr {
    AUnparseable(code:String);
    
    AName(name:String);
    
    AInt(num:Int);
    AFloat(num:Float);
    AString(str:String);
    AArrayIndex(a:AExpr, i:AExpr);
    AFieldAccess(name:String, field:String);
    AParenthesis(expr:AExpr);
    
    ANew(type:TExpr, arguments:Array<AExpr>);
    AFunctionCall(name:String, arguments:Array<AExpr>);
    AMethodCall(name:String, method:String, arguments:Array<AExpr>);

    ABlock(exprs:Array<AExpr>);
    AFor(name:String, range:AExpr, body:AExpr); // NOTE: Haxe syntax doesn't include the type of this var
    AIf(test:AExpr, ifTrue:AExpr, ifFalse:Null<AExpr>);
    AVar(name:String, type:TExpr, init:AExpr);
    AReturn(expr:AExpr);
}

// Class definition components
enum DExpr {
    DUnparseable(code:String);
    DPrecondition(test:AExpr, message:AExpr);
    DField(name:String, type:TExpr);
    DStaticVar(name:String, type:TExpr, init:AExpr);
    DMethod(name:String, type:TExpr, params:Array<{name:String, type:TExpr}>, body:AExpr);
}


class Ast {
    // Convert a Haxe expression into an AST, with a twist.
    //
    // The Haxe expression is known at compile time. I want an AST at
    // run time. So what I'm actually doing here is converting a Haxe
    // expression into Haxe expressions that generate the converted
    // AST.
    macro static public function classdef(e:Expr):Expr {
        // Convert Haxe's binary and unary op types (enum) into a
        // string, since I'm representing them as function calls even
        // if they're operators. I only convert the operators I expect
        // to use for the hex algorithms; this isn't a complete
        // translator.
        function parseBinop(op) {
            return switch(op) {
                case OpAdd: "+";
                case OpSub: "-";
                case OpMult: "*";
                case OpDiv: "/";
                case OpEq: "==";
                case OpNotEq: "!=";
                case OpGt: ">";
                case OpInterval: "#range";
                case OpAnd: "&";
                case OpBoolAnd: "and";
                case OpAssign: "set!";
                case _: return "#unknown-binop";
            };
        }
        function parseUnOp(op) {
            return switch(op) {
                case OpNot: "#not";
                case OpNeg: "#negate";
                case _: return "#unknown-unop";
            };
        }
        
        // Recursively convert a Haxe type expression into Haxe code
        // that will generate the corresponding TExpr. I handle only a
        // small set of types.
        function T(t:ComplexType):Expr {
            return switch (t) {
                case TPath({name:atom, pack:[], params:[]}):
                    switch (atom) {
                        case "Void": macro TVoid;
                        case "Int": macro TInt;
                        case "Float": macro TFloat;
                        case "String": macro TString;
                        case "Hex": macro THex;
                        case "FractionalHex": macro TFractionalHex;
                        case "OffsetCoord": macro TOffsetCoord;
                        case "DoubledCoord": macro TDoubledCoord;
                        case "Point": macro TPoint;
                        case "Layout": macro TLayout;
                        case "Orientation": macro TOrientation;
                        case _: trace("Invalid atomic type " + atom); macro TVoid;
                    }
                case TPath({name:"Array", pack:_, params:[TPType(p)]}): macro TArray(${T(p)});
                case _: trace("Haxe type not allowed: " + t); macro TVoid;
            };
        }
        
        // Recursively convert a Haxe expression into Haxe code that
        // will generate the corresponding AExpr AST. I handle only a
        // subset of Haxe code.
        function A(e:Expr):Expr {
            return switch (e.expr) {
                case EConst(CIdent(name)): macro AName($v{name});
                case EConst(CInt(n)): macro AInt($v{Std.parseInt(n)});
                case EConst(CFloat(n)): macro AFloat($v{Std.parseFloat(n)});
                case EConst(CString(s)): macro AString($v{s});
                case EArray(a, i): macro AArrayIndex(${A(a)}, ${A(i)});
                case EField({expr: EConst(CIdent(name)), pos:_}, field): macro AFieldAccess($v{name}, $v{field});
                case EUnop(op, false, expr): macro AFunctionCall($v{parseUnOp(op)}, [${A(expr)}]);
                case EBinop(op, left, right): macro AFunctionCall($v{parseBinop(op)}, [${A(left)}, ${A(right)}]);
                case EParenthesis(expr): macro AParenthesis(${A(expr)});
                case ENew(typePath, arguments): var as = [for (a in arguments) A(a)]; macro ANew(${T(TPath(typePath))}, [$a{as}]);
                case ECall({expr: EConst(CIdent(name)), pos:_}, arguments): var as = [for (a in arguments) A(a)]; macro AFunctionCall($v{name}, [$a{as}]);
                case ECall({expr: EField({expr: EConst(CIdent(name)), pos:_}, method), pos:_}, arguments): var as = [for (a in arguments) A(a)]; macro AMethodCall($v{name}, $v{method}, [$a{as}]);
                case EBlock(exprs): var as = [for (e in exprs) A(e)]; macro ABlock([$a{as}]);
                case EFor({expr: EBinop(OpIn, {expr:EConst(CIdent(name)),pos:_}, range), pos:_}, body): macro AFor($v{name}, ${A(range)}, ${A(body)});
                case EIf(test, ifTrue, null): macro AIf(${A(test)}, ${A(ifTrue)}, ABlock([]));
                case EIf(test, ifTrue, ifFalse): macro AIf(${A(test)}, ${A(ifTrue)}, ${A(ifFalse)});
                case EVars([{name:name, type:type, expr:expr}]): macro AVar($v{name}, ${T(type)}, ${A(expr)});
                case EReturn(expr): macro AReturn(${A(expr)});
                case _: var code = "" + e.expr; trace("Haxe code not allowed: "  + code); macro AUnparseable($v{code});
            };
        }

        // Top-level declarations
        function D(e:Expr) {
            return switch (e.expr) {
                case ECall({expr: EConst(CIdent("PRECONDITION")), pos:_}, [condition, message]):
                    macro DPrecondition(${A(condition)}, ${A(message)});
                case EFunction(FNamed(name, false), {args:args, expr:body, params:[], ret:type}):
                    var ps = [for (param in args) { var name = param.name, type = param.type; macro {name:$v{name}, type:${T(type)}}}];
                    macro DMethod($v{name}, ${T(type)}, [$a{ps}], ${A(body)});
                case EVars([{name:name, type:type, expr:expr}]):
                    if (expr == null) {
                        macro DField($v{name}, ${T(type)});
                    } else {
                        macro DStaticVar($v{name}, ${T(type)}, ${A(expr)});
                    }
                case _:
                    trace("classdef declaration not recognized " + e);
                    macro DUnparseable($v{e.expr});
            };
        }

        return switch (e.expr) {
            case EBlock(decls):
                var ds = [for (d in decls) D(d)];
                macro [$a{ds}];
            case _:
                trace("classdef should get a block, not " + e);
                macro [];
        };
    }
}
