// From http://www.redblobgames.com/
// Copyright 2015 Red Blob Games <redblobgames@gmail.com>
// License: Apache v2.0 <http://www.apache.org/licenses/LICENSE-2.0.html>

//
// Java output

using Ast;
using StringTools;
import Output.*;

class OutputJava extends Output {
    
    function new() {}

    // Rename functions Hex.hex_* to Hex.*, etc.
    function rename(context:String, name:String, isMethod:Bool) {
        var prefix = context.toLowerCase() + "_";
        if (name.startsWith(prefix)) { name = name.substr(prefix.length); }
        if (isMethod) { name = Output.nameToCamelCase(name); }
        return name;
    }

    // Build up a mapping of all the renames, so that when emitting
    // function calls we know what those functions are now called and
    // how to now call them. The localThisVar applies to a single method,
    // remembering which name in the original code has turned into 'this'.
    var nameMapping:Map<String, {context:String, name:String}> = new Map();
    var localThisVar:String = "";
    var instanceMethods:Map<String, Bool> = new Map();
    function collectNames(className:String, decls:Array<DExpr>) {
        for (decl in decls) {
            switch(decl) {
                case DStaticVar(name, _, _):
                    nameMapping[name] = {context: className, name: rename(className, name, false)};
                case DMethod(name, _, params, _):
                    nameMapping[name] = {context: className, name: rename(className, name, true)};
                    instanceMethods[name] = (params.length > 0 && Type(params[0].type) == className);
                case _:
            }
        }
        
        return "";
    }

    
    override function Type(t:TExpr) {
        return switch (t) {
            case TVoid: 'void';
            case TInt: 'int';
            case TFloat: 'double';
            case TArray(p): 'ArrayList<${Type(p)}>';
            case _: return super.Type(t);
        };
    }
        
    
    override function Expr(e:AExpr) {
        return switch (e) {
            case AName(name):
                if (name == "PI") "Math.PI"
                else if (name == localThisVar) 'this'
                else if (nameMapping.exists(name)) '${nameMapping[name].context}.${nameMapping[name].name}'
                else name;
            case AFieldAccess(a, i):
                if (a == localThisVar) '$i' else '$a.$i';
            case AArrayIndex(a, i): '${Expr(a)}.get(${Expr(i)})';
            case ANew(TArray(type), arguments): 'new ${Type(TArray(type))}(){{${[for (a in arguments) "add("+Expr(a)+");"].join(" ")}}}';
            case ANew(type, arguments): 'new ${Type(type)}(${[for (a in arguments) Expr(a)].join(", ")})';
            case AFunctionCall("#negate", [expr]): '-${Expr(expr)}';
            case AFunctionCall("#not", [expr]): '!${Expr(expr)}';
            case AFunctionCall("$error", [expr]): 'throw new IllegalArgumentException(${Expr(expr)})';
            case AFunctionCall("$array_length", [AName(a)]): Expr(AFieldAccess(a, "size()"));
            case AFunctionCall("Float", [expr]): Expr(expr);
            case AFunctionCall(name, arguments):
                if ((name == ">" || name == "==" || name == "!=" || name == "+" || name == "-" || name == "*" || name == "/" || name == "and" || name == "&") && arguments.length == 2) {
                    if (name == "and") { name = "&&"; }
                    '${Expr(arguments[0])} $name ${Expr(arguments[1])}';
                } else {
                    if (name == "Int") { name = "(int)"; }
                    if (name == "sin") { name = "Math.sin"; }
                    if (name == "cos") { name = "Math.cos"; }
                    if (name == "round") { name = "Math.round"; }
                    if (name == "abs") { name = "Math.abs"; }
                    if (name == "sqrt") { name = "Math.sqrt"; }
                    if (name == "max") { name = "Math.max"; }
                    if (instanceMethods[name]) {
                        name = '${Expr(arguments[0])}.${nameMapping[name].name}';
                        if (name.substr(0, 5) == 'this.') { name = name.substr(5); }
                        arguments = arguments.slice(1);
                    } else if (nameMapping.exists(name)) {
                        name = '${nameMapping[name].context}.${nameMapping[name].name}';
                    }
                    '$name(${[for (a in arguments) Expr(a)].join(", ")})';
                }
            case AMethodCall(name, method, arguments):
                if (method == "push") { method = "add"; }
                '$name.$method(${[for (a in arguments) Expr(a)].join(", ")})';
            case _: super.Expr(e);
        };
    }

                           
    function Stmt(level:Int, e:AExpr) {
        return switch (e) {
            case ABlock(exprs): '${indent(level-1)}{\n' + [for (e in exprs) Stmt(level, e)].join("") + '${indent(level-1)}}\n';
            case AFor(name, AFunctionCall("#range", [begin, AFunctionCall("+", [end, AInt(1)])]), body): '${indent(level)}for (int $name = ${Expr(begin)}; $name <= ${Expr(end)}; $name++)\n${Stmt(1+level, body)}';
            case AFor(name, AFunctionCall("#range", [begin, end]), body): '${indent(level)}for (int $name = ${Expr(begin)}; $name < ${Expr(end)}; $name++)\n${Stmt(1+level, body)}';
            case AFor(name, range, body): '${indent(level)}for (Object $name : ${Expr(range)})\n${Stmt(1+level, body)}';
            case AIf(test, ifTrue, ABlock([])): '${indent(level)}if (${Expr(test)})\n${Stmt(1+level, ifTrue)}';
            case AIf(test, ifTrue, ifFalse): '${indent(level)}if (${Expr(test)})\n${Stmt(1+level, ifTrue)}${indent(level)}else\n${Stmt(1+level, ifFalse)}';
            case AFunctionCall("set!", [AName(name), expr]): '${indent(level)}$name = ${Expr(expr)};\n';
            case AFunctionCall("set!", [AArrayIndex(AName(name), index), expr]): '${indent(level)}$name[${Expr(index)}] = ${Expr(expr)};\n';
            case AVar(name, type, init):
                var expr = Expr(init);
                if (name != expr) {
                    '${indent(level)}${Type(type)} $name = ${expr};\n';
                } else {
                    /* If it was originally var foo = hex.foo
                    and that got turned into var foo = this.foo for instance methods
                    and that got turned into var foo = foo because Java does not
                    need "this." then there's no need for this variable */
                    '';
                };
            case AReturn(expr): '${indent(level)}return ${Expr(expr)};\n';
            case _: '${indent(level)}${Expr(e)};\n';
        };
    }

                                      
    function Decl(context:String, decl:DExpr) {
        return switch(decl) {
            case DField(name, type):
                '${indent(1)}public final ${Type(type)} $name;\n';
            case DStaticVar(name, type, init):
                '${indent(1)}static public ${Type(type)} ${rename(context, name, false)} = ${Expr(init)};\n';
            case DMethod(name, type, params, body):
                var staticModifier = 'static ';
                localThisVar = '';
                if (instanceMethods[name]) {
                    staticModifier = '';
                    localThisVar = params[0].name;
                    params = params.slice(1);
                }
                '\n${indent(1)}${staticModifier}public ${Type(type)} ${rename(context, name, true)}(${[for (p in params) Type(p.type) + " " + p.name].join(", ")})\n${Stmt(2, body)}\n';
            case _: "";
        };
    }

                              
    function output(className:String, decls:Array<DExpr>) {
        var fields = collectFields(decls);
        var preconditions = collectPreconditions(decls);
        var access = className == "Tests" ? "public " : "";
        return [
            '${access}class ${className}\n{\n',
            '${indent(1)}public ${className}(',
            [for (f in fields) '${Type(f.type)} ${f.name}'].join(", "),
            ")\n",
            '${indent(1)}{\n',
            [for (f in fields) '${indent(2)}this.${f.name} = ${f.name};\n'].join(""),
            (preconditions != null)? '${indent(2)}if (${Expr(preconditions.test)}) throw new IllegalArgumentException(${Expr(preconditions.message)});\n' : '',
            '${indent(1)}}\n'
        ].join("")
             + [for (decl in decls) Decl(className, decl)].join("")
             + "}\n";
    }
    

    public static function main() {
        var output = new OutputJava();
        
        Sys.println('// Generated code -- CC0 -- No Rights Reserved -- http://www.redblobgames.com/grids/hexagons/

import java.util.ArrayList;
import java.util.Arrays;

');

        Output.outputAll(output.collectNames, false);
        Output.outputAll(output.output, false);

        Sys.println('

// Tests

');

        var testsClass = Codegen.testsClass();
        testsClass.push(DMethod("main", TVoid, [], ABlock([AFunctionCall("test_all", [])])));
        testsClass.push(DMethod("complain", TVoid, [{name: "name", type: TString}], ABlock([AFunctionCall("System.out.println", [AFunctionCall("+", [AString("FAIL "), AName("name")])])])));
        output.collectNames('Tests', testsClass);
        Sys.println(output.output('Tests', testsClass));
    }
}
