// From http://www.redblobgames.com/grids/hexagons/
// Copyright 2015 Red Blob Games <redblobgames@gmail.com>
// License: Apache v2.0 <http://www.apache.org/licenses/LICENSE-2.0.html>
//
// Rust output

using Ast;
import Output.*;

class OutputRust extends Output {
    function new() {}

    var structFieldNames: Map<String, Array<String>> = new Map();
    
    override function Type(t:TExpr) {
        return switch (t) {
            case TVoid: 'void';
            case TInt: 'i32';
            case TFloat: 'f64';
            case TString: '&str';
            case TArray(p): 'Vec<${Type(p)}>';
            case _: super.Type(t);
        };
    }

    /* added parentheses if needed for ___ as T */
    function AsExpr(e:AExpr) {
        /* NOTE: this isn't quite safe for all code but it is 
           good enough for the hex code */
        return switch (e) {
            case AFunctionCall(_, _): Expr(e);
            default: ParenExpr(e);
        }
    }
    
    /* add parentheses if needed for ___.method() */
    function ParenExpr(e:AExpr) {
        return switch (e) {
            case AName(_): Expr(e);
            case AInt(_): Expr(e);
            case AFloat(_): Expr(e);
            case AString(_): Expr(e);
            case AFieldAccess(a, i): Expr(e);
            case AParenthesis(expr): Expr(e);
            case AArrayIndex(_, _): Expr(e);
            case _: '(${Expr(e)})';
        }
    }

    
    override function Expr(e:AExpr) {
        return switch (e) {
            case AName(name): name;
            case ANew(TArray(_), arguments): 'vec![${[for (a in arguments) Expr(a)].join(", ")}]';
            case ANew(type, arguments): '${Type(type)} { ${[for (i in 0...arguments.length) structFieldNames[Type(type)][i] + ": " + Expr(arguments[i])].join(", ")} }';
            case AArrayIndex(a, i): '${Expr(a)}[${Expr(i)} as usize]';
            case AFunctionCall("$error", [expr]): 'panic!(${Expr(expr)})';
            case AFunctionCall("#negate", [expr]): '-${Expr(expr)}';
            case AFunctionCall("#not", [expr]): '!${Expr(expr)}';
            case AFunctionCall("abs", [expr]): '${ParenExpr(expr)}.abs()';
            case AFunctionCall("round", [expr]): '${ParenExpr(expr)}.round()';
            case AFunctionCall("sqrt", [AFloat(3)]): 'SQRT_3';
            case AFunctionCall("sqrt", [AFloat(num)]): '${Math.sqrt(num)}f64';
            case AFunctionCall("sqrt", [expr]): '${ParenExpr(expr)}.sqrt()';
            case AFunctionCall("sin", [expr]): '${ParenExpr(expr)}.sin()';
            case AFunctionCall("cos", [expr]): '${ParenExpr(expr)}.cos()';
            case AFunctionCall("$array_length", [AName(a)]): AsExpr(AFieldAccess(a, "len()")) + ' as i32';
            case AFunctionCall("Int", [expr]): '${AsExpr(expr)} as i32';
            case AFunctionCall("Float", [expr]): '${AsExpr(expr)} as f64';
            case AFunctionCall(name, arguments):
                if ((name == ">" || name == "==" || name == "!=" || name == "+" || name == "-" || name == "*" || name == "/" || name == "and" || name == "&") && arguments.length == 2) {
                    if (name == "and") { name = "&&"; }
                    '${Expr(arguments[0])} $name ${Expr(arguments[1])}';
                } else {
                    '$name(${[for (a in arguments) Expr(a)].join(", ")})';
                }
            case AMethodCall(name, method, arguments):
                '$name.$method(${[for (a in arguments) Expr(a)].join(", ")})';
            case _: super.Expr(e);
        };
    }
                                                         

    function Stmt(level:Int, e:AExpr) {
        return switch (e) {
            case ABlock(exprs): '${indent(level-1)}{\n' + [for (e in exprs) Stmt(level, e)].join("") + '${indent(level-1)}}\n';
            case AFor(name, AFunctionCall("#range", [begin, AFunctionCall("+", [end, AInt(1)])]), body): '${indent(level)}for $name in ${Expr(begin)} ..= ${Expr(end)}\n${Stmt(1+level, body)}';
            case AFor(name, AFunctionCall("#range", [begin, end]), body): '${indent(level)}for $name in ${Expr(begin)} .. (${Expr(end)})\n${Stmt(1+level, body)}';
            case AFor(name, range, body): '${indent(level)}for $name in ${Expr(range)}\n${Stmt(1+level, body)}';
            case AIf(test, ifTrue, ABlock([])): '${indent(level)}if ${Expr(test)}\n${Stmt(1+level, ifTrue)}';
            case AIf(test, ifTrue, ifFalse): '${indent(level)}if ${Expr(test)}\n${Stmt(1+level, ifTrue)}${indent(level)}else\n${Stmt(1+level, ifFalse)}';
            case AFunctionCall("set!", [AName(name), expr]): '${indent(level)}$name = ${Expr(expr)};\n';
            case AFunctionCall("set!", [AArrayIndex(AName(name), index), expr]): '${indent(level)}$name[${Expr(index)}] = ${Expr(expr)};\n';
            case AVar(name, type, init): '${indent(level)}let mut $name: ${Type(type)} = ${Expr(init)};\n';
            case AReturn(expr): '${indent(level)}return ${Expr(expr)};\n';
            case _: '${indent(level)}${Expr(e)};\n';
        };
    }

    
    function Decl(decl:DExpr) {
        return switch(decl) {
            case DMethod(name, type, params, body):
                'pub fn $name(${[for (p in params) p.name + ": " + Type(p.type)].join(", ")})${Type(type) == "void"? "" : "-> " + Type(type)}\n${Stmt(1, body)}\n\n';
            case DStaticVar(name, TArray(type), ANew(TArray(_), init)):
                'static $name: [${Type(type)}; ${init.length}] = [ ${[for (e in init) Expr(e)].join(", ")} ];\n';
            case DStaticVar(name, type, init):
                'const $name: ${Type(type)} = ${Expr(init)};\n';
            case _:
                ''; // fields handled elsewhere
        };
    }

                              
    function outputHeader(classname:String, decls:Array<DExpr>) {
        var output = "";
        var fields = collectFields(decls);
        var preconditions = collectPreconditions(decls);
        if (fields.length > 0) {
            output += '#[derive(Clone, Copy, Debug)]\npub struct ${classname}\n{\n';
            structFieldNames[classname] = [];
            for (f in fields) {
                output += '${indent(1)}pub ${f.name}: ${Type(f.type)},\n';
                structFieldNames[classname].push(f.name);
            }
            output += "}\n\n";
        }
        return output;
    }
    

    function outputDefinition(classname:String, decls:Array<DExpr>) {
        return [for (decl in decls) Decl(decl)].join("");
    }

    
    static function main() {
        var output = new OutputRust();
        
        Sys.println('// Generated code -- CC0 -- No Rights Reserved -- http://www.redblobgames.com/grids/hexagons/

#![allow(non_snake_case)]
#![allow(non_upper_case_globals)]
#![allow(unused_mut)] // TODO: remove this by analyzing output code

use std::cmp::max;
use std::f64::consts::PI;
const SQRT_3: f64 = ${Math.sqrt(3)};
');
        
        Output.outputAll(output.outputHeader, false);
        
        Output.outputAll(output.outputDefinition, false);

        Sys.println('
// Tests

fn complain(name: &str)
{
  print!("FAIL {}\\n", name);
}
');

        Output.outputTests(output.outputHeader);
        Output.outputTests(output.outputDefinition);
        
        Sys.println('

fn main() {
  test_all();
}
');
        
    }
}

